searchData={"items":[{"type":"module","title":"Antenna","doc":"`Antenna` is a mixture of [`Phoenix.PubSub`](https://hexdocs.pm/phoenix_pubsub/Phoenix.PubSub.html)\n  and [`:gen_event`](https://www.erlang.org/doc/apps/stdlib/gen_event.html) functionality\n  with some batteries included.\n\nIt implements back-pressure on top of `GenStage`, is fully conformant with\n  [OTP Design Principles](https://www.erlang.org/doc/system/events). and\n  is distributed out of the box.\n\n`Antenna` supports both asynchronous _and_ synchronous events. While the most preferrable way\n  would be to stay fully async with `Antenna.event/3`, one still might propagate the event\n  synchronously with `Antenna.sync_event/3` and collect all the responses from all the handlers.\n\nOne can have as many isolated `Antenna`s as necessary, distinguished by `Antenna.t:id/0`.","ref":"Antenna.html"},{"type":"module","title":"Getting Started - Antenna","doc":"To start using Antenna, add it to your dependencies in mix.exs:\n\n```elixir\ndef deps do\n  [\n    {:antenna, \"~> 0.3.0\"}\n  ]\nend\n```\n\nThen add it to your supervision tree:\n\n```elixir\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [\n      Antenna\n    ]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\nend\n```","ref":"Antenna.html#module-getting-started"},{"type":"module","title":"Configuration - Antenna","doc":"Antenna can be configured in your config.exs:\n\n```elixir\nconfig :antenna,\n  id: MyApp.Antenna,           # Optional custom identifier\n  distributed: true,           # Enable distributed mode (default: true)\n  sync_timeout: 5_000          # Default timeout for sync_event/4\n```","ref":"Antenna.html#module-configuration"},{"type":"module","title":"Sequence Diagram - Antenna","doc":"```mermaid\nsequenceDiagram\n  Consumer->>+Broadcaster: sync_event(channel, event)\n  Consumer->>+Broadcaster: event(channel, event) \n  Broadcaster-->>+Consumer@Node1: event\n  Broadcaster-->>+Consumer@Node2: event\n  Broadcaster-->>+Consumer@NodeN: event\n  Consumer@Node1-->>-NoOp: mine?\n  Consumer@NodeN-->>-NoOp: mine?\n  Consumer@Node2-->>+Matchers: event\n  Matchers-->>+Handlers: handle_match/2\n  Matchers-->>+Handlers: handle_match/2\n  Handlers->>-Consumer: response(to: sync_event)\n```\n\n[ASCII representation](https://cascii.app/4164d).","ref":"Antenna.html#module-sequence-diagram"},{"type":"module","title":"Common Use Cases - Antenna","doc":"","ref":"Antenna.html#module-common-use-cases"},{"type":"module","title":"Basic Event Handling - Antenna","doc":"```elixir\n# Define a matcher for specific events\nAntenna.match(MyApp.Antenna, {:user_registered, user_id, _meta}, fn channel, event ->\n  # Handle the event\n  {:user_registered, user_id, meta} = event\n  Logger.info(\"New user registered: #{user_id} on channel #{channel}\")\n  MyApp.UserNotifier.welcome(user_id)\nend, channels: [:user_events])\n\n# Send events\nAntenna.event(MyApp.Antenna, [:user_events], {:user_registered, \"user123\", %{source: \"web\"}})\n```","ref":"Antenna.html#module-basic-event-handling"},{"type":"module","title":"Error Handling - Antenna","doc":"```elixir\n# Match on error events\nAntenna.match(MyApp.Antenna, {:error, reason, _context} = event, fn channel, event ->\n  # Log and handle errors\n  Logger.error(\"Error on channel #{channel}: #{inspect(event)}\")\n  MyApp.ErrorTracker.track(event)\nend, channels: [:errors])\n\n# Send error events\nAntenna.event(MyApp.Antenna, [:errors], {:error, :validation_failed, %{user_id: 123}})\n```","ref":"Antenna.html#module-error-handling"},{"type":"module","title":"Distributed Setup - Antenna","doc":"Antenna automatically handles distribution when nodes are connected:\n\n```elixir\n# On node1@host1\nNode.connect(:\"node2@host2\")\nAntenna.match(MyApp.Antenna, {:replicate, data}, fn _, {:replicate, data} ->\n  MyApp.Storage.replicate(data)\nend, channels: [:replication])\n\n# On node2@host2\nAntenna.event(MyApp.Antenna, [:replication], {:replicate, some_data})\n# The event will be processed on both nodes\n```","ref":"Antenna.html#module-distributed-setup"},{"type":"module","title":"Custom Matchers - Antenna","doc":"You can use pattern matching and guards in your matchers:\n\n```elixir\n# Match on specific patterns with guards\nAntenna.match(MyApp.Antenna, \n  {:temperature, celsius, _} when celsius > 30,\n  fn _, {:temperature, c, location} ->\n    Logger.warning(\"High temperature (#{c}°C) detected at #{location}\")\n  end,\n  channels: [:sensors])\n\n# Match on map patterns\nAntenna.match(MyApp.Antenna,\n  %{event: :payment, amount: amount} when amount > 1000,\n  fn _, event ->\n    MyApp.LargePaymentProcessor.handle(event)\n  end,\n  channels: [:payments])\n```","ref":"Antenna.html#module-custom-matchers"},{"type":"module","title":"Usage Example - Antenna","doc":"The consumer of this library is supposed to declare one or more matchers, subscribing to one\n  or more channels, and then call `Antenna.event/2` to propagate the event.\n\n```elixir\nassert {:ok, _pid, \"{:tag_1, a, _} when is_nil(a)\"} =\n  Antenna.match(Antenna, {:tag_1, a, _} when is_nil(a), self(), channels: [:chan_1])\n\nassert :ok = Antenna.event(Antenna, [:chan_1], {:tag_1, nil, 42})\nassert_receive {:antenna_event, :chan_1, {:tag_1, nil, 42}}\n```","ref":"Antenna.html#module-usage-example"},{"type":"module","title":"Architecture - Antenna","doc":"Antenna uses a distributed pub/sub architecture with the following components:\n\n1. **Broadcaster** - Handles event distribution across nodes\n2. **Matchers** - Pattern match on events and delegate to handlers\n3. **Guard** - Manages channel subscriptions and handler registration\n4. **PubSub** - Implements the core pub/sub functionality\n\nEvents flow through the system as follows:\n\n1. Client calls `event/3` or `sync_event/4`\n2. Broadcaster distributes the event to all nodes\n3. Matchers on each node pattern match against the event\n4. Matching handlers process the event\n5. For sync_event, responses are collected and returned","ref":"Antenna.html#module-architecture"},{"type":"module","title":"Best Practices - Antenna","doc":"1. **Channel Organization**\n   - Use atoms for channel names\n   - Group related events under common channels\n   - Consider using hierarchical channel names (e.g., :users_created, :users_updated)\n\n2. **Pattern Matching**\n   - Use specific patterns to avoid unnecessary pattern matches\n   - Include guards for more precise matching\n   - Consider the order of pattern matches when using multiple matchers\n\n3. **Handler Design**\n   - Keep handlers focused and single-purpose\n   - Use `sync_event/4` when you need responses\n   - Consider timeouts for sync operations\n   - Handle errors within handlers to prevent cascade failures\n\n4. **Performance**\n   - Use async events (`event/3`) when possible\n   - Keep handler processing time minimal\n   - Consider using separate processes for long-running operations\n   - Monitor matcher and handler counts\n\n5. **Testing**\n   - Test matchers with various event patterns\n   - Verify handler behavior\n   - Test distributed scenarios\n   - Use ExUnit's async: true when possible","ref":"Antenna.html#module-best-practices"},{"type":"type","title":"Antenna.channel/0","doc":"The identifier of the channel, messages can be sent to, preferrably `atom()`","ref":"Antenna.html#t:channel/0"},{"type":"function","title":"Antenna.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Antenna.html#child_spec/1"},{"type":"type","title":"Antenna.event/0","doc":"The event being sent to the listeners","ref":"Antenna.html#t:event/0"},{"type":"function","title":"Antenna.event/3","doc":"Sends an event to all the associated matchers through channels.\n\nThe special `:*` might be specified as channels, then the event\n  will be sent to all the registered channels.\n\nIf one wants to collect results of all the registered event handlers,\n  they should look at `sync_event/3` instead.","ref":"Antenna.html#event/3"},{"type":"function","title":"Examples - Antenna.event/3","doc":"","ref":"Antenna.html#event/3-examples"},{"type":"function","title":"Basic Usage - Antenna.event/3","doc":"```elixir\n# Send event to a single channel\nAntenna.event(MyApp.Antenna, :user_events, {:user_logged_in, \"user123\"})\n\n# Send event to multiple channels\nAntenna.event(MyApp.Antenna, [:logs, :metrics], {:api_call, \"/users\", 200, 45})\n\n# Send event to all channels\nAntenna.event(MyApp.Antenna, :*, {:system_notification, :service_starting})\n```","ref":"Antenna.html#event/3-basic-usage"},{"type":"function","title":"Custom Antenna ID - Antenna.event/3","doc":"```elixir\n# Using a specific Antenna instance\nAntenna.event(MyBackgroundJobs.Antenna, :jobs, {:job_completed, \"job-123\"})\n```","ref":"Antenna.html#event/3-custom-antenna-id"},{"type":"function","title":"Antenna.handle/3","doc":"Adds a handler to the matcher process specified by `pid`","ref":"Antenna.html#handle/3"},{"type":"type","title":"Antenna.handler/0","doc":"The actual handler to be associated with an event(s). It might be either a function\n  or a process id, in which case the message of a following shape will be sent to it.\n\n```elixir\n{:antenna_event, channel, event}\n```","ref":"Antenna.html#t:handler/0"},{"type":"type","title":"Antenna.id/0","doc":"The identifier of the isolated `Antenna`","ref":"Antenna.html#t:id/0"},{"type":"macro","title":"Antenna.match/4","doc":"Declares a matcher for tagged events.\n\nThis function sets up a pattern matcher to handle specific events on the given channels.\nWhen an event matching the pattern is sent to any of the channels, the provided handler\nwill be invoked.","ref":"Antenna.html#match/4"},{"type":"macro","title":"Options - Antenna.match/4","doc":"* `:channels` - A list of channels to subscribe this matcher to\n* `:once?` - If true, the matcher will be automatically unmatched after the first match (default: false)","ref":"Antenna.html#match/4-options"},{"type":"macro","title":"Examples - Antenna.match/4","doc":"","ref":"Antenna.html#match/4-examples"},{"type":"macro","title":"Basic Usage - Antenna.match/4","doc":"```elixir\nAntenna.match(Antenna, %{tag: _, success: false}, fn channel, message ->\n  Logger.warning(\"The processing failed for [\" <> \n    inspect(channel) <> \"], result: \" <> inspect(message))\nend, channels: [:rabbit])\n```","ref":"Antenna.html#match/4-basic-usage"},{"type":"macro","title":"One-time Matcher - Antenna.match/4","doc":"```elixir\n# This matcher will only trigger once and then be removed\nAntenna.match(Antenna, {:init, pid}, fn _, {:init, pid} ->\n  send(pid, :initialized)\nend, channels: [:system], once?: true)\n```","ref":"Antenna.html#match/4-one-time-matcher"},{"type":"macro","title":"Using Pattern Guards - Antenna.match/4","doc":"```elixir\nAntenna.match(Antenna, \n  {:metric, name, value} when value > 100, \n  fn channel, event ->\n    Logger.info(\"High metric value on #{channel}: #{inspect(event)}\")\n  end, \n  channels: [:metrics])\n```","ref":"Antenna.html#match/4-using-pattern-guards"},{"type":"macro","title":"Multiple Channels - Antenna.match/4","doc":"```elixir\nAntenna.match(Antenna, \n  {:user_event, user_id, _action},\n  fn channel, event ->\n    # Handle user events from multiple channels\n    MyApp.UserEventTracker.track(channel, event)\n  end,\n  channels: [:user_logins, :user_actions, :user_settings])\n```","ref":"Antenna.html#match/4-multiple-channels"},{"type":"type","title":"Antenna.matcher/0","doc":"The matcher to be used for the event, e.g. `{:tag_1, a, _} when is_nil(a)`\n\nThe pattern matching is done on the `Antenna` process, so it might be\n  either a function or a process id.","ref":"Antenna.html#t:matcher/0"},{"type":"type","title":"Antenna.opts/0","doc":"The options to be passed to `Antenna.match/4` and `Antenna.unmatch/2`\n\n* `:id` - the identifier of the `Antenna`, defaults to `Antenna`\n* `:channels` - a list of channels to subscribe the matcher to\n* `:once?` - if true, the matcher would be removed after the first match","ref":"Antenna.html#t:opts/0"},{"type":"function","title":"Antenna.registered_matchers/1","doc":"Returns a map of matches to matchers","ref":"Antenna.html#registered_matchers/1"},{"type":"function","title":"Antenna.start_link/1","doc":"Starts the `Antenna` matcher tree for the `id` given.","ref":"Antenna.html#start_link/1"},{"type":"function","title":"Antenna.subscribe/3","doc":"Subscribes a matcher process specified by `pid` to a channel(s)","ref":"Antenna.html#subscribe/3"},{"type":"function","title":"Antenna.sync_event/5","doc":"Sends an event to all the associated matchers through channels synchronously,\ncollecting and returning responses from all handlers.\n\nThe special `:*` might be specified as channels, then the event\nwill be sent to all the registered channels.\n\nUnlike `event/3`, this function waits for all handlers to process the event\nand returns their responses.","ref":"Antenna.html#sync_event/5"},{"type":"function","title":"Parameters - Antenna.sync_event/5","doc":"* `id` - The Antenna instance identifier (optional, defaults to configured value)\n* `channels` - Single channel or list of channels to send the event to\n* `event` - The event to be sent\n* `timeout` - Maximum time to wait for responses (default: 5000ms)","ref":"Antenna.html#sync_event/5-parameters"},{"type":"function","title":"Examples - Antenna.sync_event/5","doc":"","ref":"Antenna.html#sync_event/5-examples"},{"type":"function","title":"Basic Usage - Antenna.sync_event/5","doc":"```elixir\n# Send sync event and collect responses\nresults = Antenna.sync_event(MyApp.Antenna, :user_events, {:verify_user, \"user123\"})\n\n# Send to multiple channels with custom timeout\nresults = Antenna.sync_event(\n  MyApp.Antenna,\n  [:validations, :security],\n  {:user_action, \"user123\", :delete_account},\n  10_000\n)\n```","ref":"Antenna.html#sync_event/5-basic-usage"},{"type":"function","title":"Response Format - Antenna.sync_event/5","doc":"The function returns a list of handler responses, each wrapped in a tuple\nindicating whether the pattern matched and including the handler's result:\n\n```elixir\n[\n  {:match, %{\n    match: \"pattern_string\",\n    channel: :channel_name,\n    results: [handler_result]\n  }},\n  {:no_match, :channel_name}\n]\n```","ref":"Antenna.html#sync_event/5-response-format"},{"type":"function","title":"Antenna.unhandle/3","doc":"Removes a handler from the matcher process specified by `pid`","ref":"Antenna.html#unhandle/3"},{"type":"macro","title":"Antenna.unmatch/2","doc":"Undeclares a matcher for tagged events previously declared with `Antenna.match/4`.\n\nAccepts both an original match _or_ a name returned by `Antenna.match/4`,\n  which is effectively `Macro.to_string(match)`.","ref":"Antenna.html#unmatch/2"},{"type":"macro","title":"Example - Antenna.unmatch/2","doc":"```elixir\nAntenna.unmatch(Antenna, %{tag: _, success: false})\n```","ref":"Antenna.html#unmatch/2-example"},{"type":"function","title":"Antenna.unsubscribe/3","doc":"Unsubscribes a previously subscribed matcher process specified by `pid` from the channel(s)","ref":"Antenna.html#unsubscribe/3"},{"type":"module","title":"Antenna.ExUnit","doc":"The helper module for ExUnit tests. It exports `assert_event/2` and `refute_event/2` functions to assert and refute events.","ref":"Antenna.ExUnit.html"},{"type":"macro","title":"Antenna.ExUnit.assert_event/6","doc":"Asserts that an event is received within the specified timeout.","ref":"Antenna.ExUnit.html#assert_event/6"},{"type":"macro","title":"Example - Antenna.ExUnit.assert_event/6","doc":"```elixir\nassert_event(Antenna, {:foo, _}, :chan_1, _, fn -> provoke_foo(Fooer) end)\n```","ref":"Antenna.ExUnit.html#assert_event/6-example"},{"type":"behaviour","title":"Antenna.Matcher","doc":"The behaviour defining matchers in the `Antenna` pub-sub model.\n\nWhile the matcher might be a simple function, or a process identifier,\n  using this behaviour would have your testing process simplified.","ref":"Antenna.Matcher.html"},{"type":"function","title":"Antenna.Matcher.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Antenna.Matcher.html#child_spec/1"},{"type":"callback","title":"Antenna.Matcher.handle_match/2","doc":"The funcion to be called back upon match","ref":"Antenna.Matcher.html#c:handle_match/2"},{"type":"type","title":"Antenna.Matcher.t/0","doc":"The module implementing `handle_match/2` callback","ref":"Antenna.Matcher.html#t:t/0"},{"type":"extras","title":"Antenna","doc":"# Antenna\n\n**The tiny framework to simplify work with events, based on `GenStage`**","ref":"readme.html"},{"type":"extras","title":"Objective - Antenna","doc":"`Antenna` is a mixture of [`Phoenix.PubSub`](https://hexdocs.pm/phoenix_pubsub/Phoenix.PubSub.html)\nand [`:gen_event`](https://www.erlang.org/doc/apps/stdlib/gen_event.html) functionality\nwith some batteries included.\n\nIt implements back-pressure on top of `GenStage`, is fully conformant with\n[OTP Design Principles](https://www.erlang.org/doc/system/events). and\nis distributed out of the box.\n\n`Antenna` supports both asynchronous _and_ synchronous events. While the most preferrable way\nwould be to stay fully async with `Antenna.event/3`, one still might propagate the event\nsynchronously with `Antenna.sync_event/3` and collect all the responses from all the handlers.\n\nOne can have as many isolated `Antenna`s as necessary, distinguished by `Antenna.t:id/0`.\n\nThe workflow looks like shown below.","ref":"readme.html#objective"},{"type":"extras","title":"Flow - Antenna","doc":"```\n                        ┌───────────────┐                                               \n                        │               │       Antenna.event(AntID, :tag1, %{foo: 42}) \n                        │  Broadcaster  │───────────────────────────────────────────────\n                        │               │                                               \n                        └───────────────┘                                               \n                              ──/──                                                     \n                        ────── /   ─────                                                \n                  ──────      /         ──────                                          \n            ──────           /                ─────                                     \n        ────                /                      ───                                  \n┌────────────────┐  ┌────────────────┐       ┌────────────────┐                         \n│                │  │                │       │                │                         \n│ Consumer@node1 │  │ Consumer@node1 │   …   │ Consumer@nodeN │                         \n│                │  │                │       │                │                         \n└────────────────┘  └────────────────┘       └────────────────┘                         \n        ·                   ·                         ·                                 \n      /   \\               /   \\                     /   \\                               \n    /       \\           /       \\                 /       \\                             \n  ·   mine?   ·       ·   mine?   ·             ·   mine?   ·                           \n    \\       /           \\       /                 \\       /                             \n      \\   /               \\   /                     \\   /                               \n        ·                   ·                         ·                                 \n        │                   │                         │                                 \n        │                   │                         │                                 \n      ─────                 │                       ─────                               \n                            │                                                           \n                            │                                                           \n                     ┌──────────────┐                                                   \n                     │              │             if (match?), do: call_handlers(event) \n                     │   matchers   │───────────────────────────────────────────────────\n                     │              │                                                   \n                     └──────────────┘                                                   \n```\n\n[ASCII representation](https://cascii.app/4164d).","ref":"readme.html#flow"},{"type":"extras","title":"Usage Example - Antenna","doc":"The consumer of this library is supposed to declare one or more matchers, subscribing to one\nor more channels, and then call `Antenna.event/2` to propagate the event.","ref":"readme.html#usage-example"},{"type":"extras","title":"Installation - Antenna","doc":"```elixir\ndef deps do\n  [\n    {:antenna, \"~> 0.2\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Is it of any good? - Antenna","doc":"Sure, it is.\n\n## [Documentation](https://hexdocs.pm/antenna)","ref":"readme.html#is-it-of-any-good"},{"type":"extras","title":"Smart Eventing","doc":"# Smart Eventing\n\nAntenna is an elegant event handling framework that combines the best features of Phoenix.PubSub and gen_event with sophisticated pattern matching and back-pressure control. Built on GenStage and fully distributed by design, it offers a powerful yet simple way to handle events across your Elixir applications.","ref":"why-antenna.html"},{"type":"extras","title":"Key Features - Smart Eventing","doc":"","ref":"why-antenna.html#key-features"},{"type":"extras","title":"1. Smart Pattern Matching - Smart Eventing","doc":"Unlike traditional pub/sub systems, Antenna provides Elixir-native pattern matching capabilities:\n\n```elixir\n# Match on complex patterns with guards\nAntenna.match(MyApp.Antenna, \n  {:temperature, celsius, location} when celsius > 30,\n  fn channel, event ->\n    Logger.warning(\"High temperature alert: #{inspect(event)} on #{channel}\")\n  end,\n  channels: [:sensors])\n```","ref":"why-antenna.html#1-smart-pattern-matching"},{"type":"extras","title":"2. Built for Distribution - Smart Eventing","doc":"- Automatic event propagation across connected nodes\n- Node-local pattern matching for efficiency\n- Built-in back-pressure through GenStage\n- Process group-based channel management\n\n```elixir\n# Events automatically propagate to all connected nodes\nNode.connect(:\"node2@host2\")\nAntenna.event(MyApp.Antenna, [:alerts], {:system_status, :degraded})\n```","ref":"why-antenna.html#2-built-for-distribution"},{"type":"extras","title":"3. Flexible Event Handling - Smart Eventing","doc":"Support for both asynchronous and synchronous event processing:\n\n```elixir\n# Async event dispatch\nAntenna.event(MyApp.Antenna, [:logs], {:user_action, user_id, :login})\n\n# Sync event with collected responses\nresponses = Antenna.sync_event(MyApp.Antenna, [:auth], {:verify_token, token})\n```","ref":"why-antenna.html#3-flexible-event-handling"},{"type":"extras","title":"4. Channel-based Organization - Smart Eventing","doc":"- Multiple isolated Antenna instances\n- Channel-based event routing\n- Support for wildcard subscriptions\n- Dynamic channel subscription management\n\n```elixir\n# Different antennas for different subsystems\nAntenna.event(:metrics, [:system], %{cpu: 80, memory: 70})\nAntenna.event(:business, [:orders], %{order_id: \"123\", status: :completed})\n```","ref":"why-antenna.html#4-channel-based-organization"},{"type":"extras","title":"5. OTP-Compliant Design - Smart Eventing","doc":"- Follows OTP design principles\n- Supervisor-based process management\n- Clean process isolation\n- Proper error handling and recovery","ref":"why-antenna.html#5-otp-compliant-design"},{"type":"extras","title":"Real-world Applications - Smart Eventing","doc":"","ref":"why-antenna.html#real-world-applications"},{"type":"extras","title":"Event-Driven Microservices - Smart Eventing","doc":"```elixir\n# Service A: Order Processing\nAntenna.match(OrderSystem, \n  %{event: :order_created, total: total} when total > 1000,\n  fn _, event ->\n    HighValueOrderProcessor.handle(event)\n  end,\n  channels: [:orders])\n\n# Service B: Notification System\nAntenna.match(NotificationSystem,\n  {:order_status, order_id, :completed},\n  fn _, event ->\n    CustomerNotifier.order_completed(order_id)\n  end,\n  channels: [:order_updates])\n```","ref":"why-antenna.html#event-driven-microservices"},{"type":"extras","title":"IoT Data Processing - Smart Eventing","doc":"```elixir\n# Device data handling with pattern matching\nAntenna.match(IoTSystem,\n  {:sensor_data, device_id, readings} when readings.temperature > 90,\n  fn channel, event ->\n    DeviceAlertHandler.process_alert(channel, event)\n  end,\n  channels: [:device_telemetry])\n```","ref":"why-antenna.html#iot-data-processing"},{"type":"extras","title":"Distributed Systems Monitoring - Smart Eventing","doc":"```elixir\n# System metrics collection\nAntenna.match(Monitoring,\n  {:metric, node_name, metric, value} when value > threshold(),\n  fn _, event ->\n    MetricsAggregator.record_threshold_breach(event)\n  end,\n  channels: [:system_metrics])\n```","ref":"why-antenna.html#distributed-systems-monitoring"},{"type":"extras","title":"Getting Started - Smart Eventing","doc":"1. Add Antenna to your dependencies:\n\n```elixir\ndef deps do\n  [{:antenna, \"~> 0.2\"}]\nend\n```\n\n2. Add it to your supervision tree:\n\n```elixir\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [\n      {Antenna, name: MyApp.EventSystem}\n    ]\n    \n    Supervisor.start_link(children, strategy: :one_for_one)\n  end\nend\n```\n\n3. Set up your event handlers:\n\n```elixir\n# Define matchers for your events\nAntenna.match(MyApp.EventSystem,\n  {:user_event, user_id, action},\n  fn channel, event ->\n    MyApp.EventTracker.track(channel, event)\n  end,\n  channels: [:user_activity])\n\n# Start sending events\nAntenna.event(MyApp.EventSystem, [:user_activity], \n  {:user_event, \"user123\", :login})\n```","ref":"why-antenna.html#getting-started"},{"type":"extras","title":"Why Choose Antenna? - Smart Eventing","doc":"1. **Native Pattern Matching**: Leverage Elixir's powerful pattern matching for event handling\n2. **Distributed by Design**: Built for distributed systems with automatic node synchronization\n3. **Back-pressure Control**: Built on GenStage for controlled event flow\n4. **Flexible Configuration**: Multiple isolated instances and channel-based routing\n5. **Production Ready**: OTP-compliant with proper supervision and error handling","ref":"why-antenna.html#why-choose-antenna"},{"type":"extras","title":"Internal Architecture - Smart Eventing","doc":"","ref":"why-antenna.html#internal-architecture"},{"type":"extras","title":"Event Flow - Smart Eventing","doc":"Antenna processes events through a distributed pipeline:\n\n```txt\n                        ┌───────────────┐                                               \n                        │               │       Antenna.event(AntID, :tag1, %{foo: 42}) \n                        │  Broadcaster  │───────────────────────────────────────────────\n                        │               │                                               \n                        └───────────────┘                                               \n                              ──/──                                                     \n                        ────── /   ─────                                                \n                  ──────      /         ──────                                          \n            ──────           /                ─────                                     \n        ────                /                      ───                                  \n┌────────────────┐  ┌────────────────┐       ┌────────────────┐                         \n│                │  │                │       │                │                         \n│ Consumer@node1 │  │ Consumer@node1 │   …   │ Consumer@nodeN │                         \n│                │  │                │       │                │                         \n└────────────────┘  └────────────────┘       └────────────────┘                         \n        ·                   ·                         ·                                 \n      /   \\               /   \\                     /   \\                               \n    /       \\           /       \\                 /       \\                             \n  ·   mine?   ·       ·   mine?   ·             ·   mine?   ·                           \n    \\       /           \\       /                 \\       /                             \n      \\   /               \\   /                     \\   /                               \n        ·                   ·                         ·                                 \n        │                   │                         │                                 \n        │                   │                         │                                 \n      ─────                 │                       ─────                               \n                            │                                                           \n                            │                                                           \n                     ┌──────────────┐                                                   \n                     │              │             if (match?), do: call_handlers(event) \n                     │   matchers   │───────────────────────────────────────────────────\n                     │              │                                                   \n                     └──────────────┘                                                   \n```","ref":"why-antenna.html#event-flow"},{"type":"extras","title":"Components - Smart Eventing","doc":"1. **Broadcaster (GenStage Producer)**\n   - Distributes events to all nodes\n   - Implements back-pressure control\n   - Handles sync/async event dispatch\n\n2. **Process Groups**\n   - Channel-based process groups for event routing\n   - Dynamic subscriber management\n   - Node-local event matching\n\n3. **Pattern Matchers**\n   - Handle complex pattern matching\n   - Support guards and conditions\n   - Route events to appropriate handlers\n\n4. **Guard**\n   - Supervises channel subscriptions\n   - Manages handler lifecycle\n   - Ensures clean process termination","ref":"why-antenna.html#components"},{"type":"extras","title":"Back-pressure - Smart Eventing","doc":"Antenna uses GenStage to implement back-pressure:\n\n- Events are dispatched based on consumer demand\n- System stays responsive under load\n- Memory usage remains stable\n- Event ordering is preserved","ref":"why-antenna.html#back-pressure"},{"type":"extras","title":"Distribution - Smart Eventing","doc":"Built-in distribution capabilities:\n\n- Automatic node discovery\n- Transparent event propagation\n- Node-local pattern matching\n- Efficient event routing\n\nFor a detailed implementation guide and more examples, visit the [official documentation](https://hexdocs.pm/antenna).","ref":"why-antenna.html#distribution"},{"type":"extras","title":"Configuration - Smart Eventing","doc":"Antenna can be configured in your `config.exs`:\n\n```elixir\nconfig :antenna,\n  id: MyApp.Antenna,           # Optional custom identifier\n  distributed: true,           # Enable distributed mode (default: true)\n  sync_timeout: 5_000          # Default timeout for sync_event/4\n```","ref":"why-antenna.html#configuration"},{"type":"extras","title":"Best Practices - Smart Eventing","doc":"","ref":"why-antenna.html#best-practices"},{"type":"extras","title":"1. Channel Organization - Smart Eventing","doc":"- Use atoms for channel names when possible\n- Group related events under common channels\n- Consider using hierarchical channel naming:\n  ```elixir\n  # Good channel organization\n  Antenna.event(MyApp.Antenna, [:users_created], event)\n  Antenna.event(MyApp.Antenna, [:users_updated], event)\n  ```","ref":"why-antenna.html#1-channel-organization"},{"type":"extras","title":"2. Pattern Matching - Smart Eventing","doc":"- Use specific patterns to avoid unnecessary matches:\n  ```elixir\n  # Good - specific pattern\n  Antenna.match(MyApp.Antenna, {:user_created, user_id, meta}, handler)\n  \n  # Less efficient - overly broad pattern\n  Antenna.match(MyApp.Antenna, {_action, _id, _}, handler)\n  ```\n- Include guards for more precise matching\n- Consider the order of pattern matches when using multiple matchers","ref":"why-antenna.html#2-pattern-matching"},{"type":"extras","title":"3. Handler Design - Smart Eventing","doc":"- Keep handlers focused and single-purpose\n- Use `sync_event/4` only when you need responses\n- Consider timeouts for sync operations:\n  ```elixir\n  # Set appropriate timeout for sync operations\n  responses = Antenna.sync_event(MyApp.Antenna, [:auth], event, 10_000)\n  ```\n- Handle errors within handlers to prevent cascade failures","ref":"why-antenna.html#3-handler-design"},{"type":"extras","title":"4. Performance - Smart Eventing","doc":"- Use async events (`event/3`) by default\n- Keep handler processing time minimal\n- Consider using separate processes for long-running operations:\n  ```elixir\n  Antenna.match(MyApp.Antenna, event_pattern, fn channel, event ->\n    # Spawn long-running operations\n    Task.start(fn -> LongRunningProcessor.process(event) end)\n  end)\n  ```\n- Monitor matcher and handler counts","ref":"why-antenna.html#4-performance"},{"type":"extras","title":"5. Testing - Smart Eventing","doc":"- Test matchers with various event patterns:\n  ```elixir\n  test \"matches high temperature events\" do\n    Antenna.match(TestAntenna, \n      {:temperature, val, _} when val > 30,\n      self(),\n      channels: [:test])\n      \n    Antenna.event(TestAntenna, [:test], {:temperature, 35, :room})\n    assert_receive {:antenna_event, :test, {:temperature, 35, :room}}\n  end\n  ```\n- Test handler behavior with both valid and invalid events\n- Test distributed scenarios\n- Use ExUnit's `async: true` when possible","ref":"why-antenna.html#5-testing"},{"type":"extras","title":"Conclusion - Smart Eventing","doc":"Antenna provides a robust foundation for building distributed event-driven systems in Elixir. Whether you're building microservices, processing IoT data, or managing system events, Antenna offers the tools you need with the elegance and reliability you expect from Elixir applications.","ref":"why-antenna.html#conclusion"},{"type":"extras","title":"DIY Message Broker","doc":"# DIY Message Broker\n\nIn Erlang (and Elixir), I always missed a way to organize “streaming” message exchange, similar to what a Message Broker provides. Normal developers accept the constraints imposed by their frameworks: Phoenix has [`PubSub`](https://hexdocs.pm/phoenix_pubsub/Phoenix.PubSub.html), OTP has [`:gen_event`](https://www.erlang.org/doc/apps/stdlib/gen_event.html), and Elixir had [`GenEvent`](https://hexdocs.pm/elixir/GenEvent.html), deprecated even before it was born.\n\nNone of these suited me. PubSub is powerful, but its asynchronicity is nailed down, and sometimes you need to call subscribers synchronously (I know you can send your pid with the message and wait for responses, but this ad-hoc approach unfortunately doesn’t help when the number of subscribers is unknown). `:gen_event` is almost what I need, but its design clearly exposes side-effect handlers, making convenient filtering of incoming message streams turn into spaghetti. In general, as always: if you don’t do it yourself, no one will.\n\nThus, the [`Antenna`](https://hexdocs.pm/antenna/) library was born, providing all the features usually achieved by adding an extra message broker dependency. My requirements for it were as follows:\n\n- **Back pressure support out of the box** (a million events sent at once should not overwhelm handlers, and adding new nodes to the cluster should transparently offload them)\n- **Channels (topics, tags)** where you can send an event that will be delivered to all registered handlers\n- **Handlers (besides channel subscriptions)** should be able to transparently filter the incoming message stream using the full power of Erlang’s pattern matching\n- **Both handlers and subscriptions** should successfully recover after a crash\n- **The message sender can request synchronous execution** of handlers, blocking the calling process until all responses are received (like RPC)\n- **The library should provide convenient primitives for testing its usage**\n\nAnyone who has used [_RabbitMQ_](https://www.rabbitmq.com/tutorials) would have noticed where I stole my ideas from easily.","ref":"diy-message-broker.html"},{"type":"extras","title":"Pseudocode - DIY Message Broker","doc":"For a long time, I’ve used a kind of [_TDD_](https://en.wikipedia.org/wiki/Test-driven_development) with a hint of [_DDD_](https://en.wikipedia.org/wiki/Domain-driven_design), or vice versa—I’m not strong in buzzword abbreviations. In general, I start designing any piece of code by calling its (yet non-existent) API. First, I make sure that calling my code is easy and pleasant, and I’ll figure out the implementation later. So I sketched something like this:\n\n```elixir\n{:ok, pid} = match({:tag_answer, _}, self())\nsubscribe(:chan, pid)\n\nevent(:chan, {:tag_answer, 42})\n\nreceive do\n  {:event, :chan, {:tag_answer, 42}} -> :ok\nafter\n  1_000 -> :error\nend\n```\n\n\nThe first two lines are handler registration and subscription. The next is sending a message. The final block checks that the message was received. These three pieces will usually be in different, unrelated parts of the code.\n\nThe first thing I noticed is that the first two lines almost always go together, so it makes sense to accept a list of channels for subscription right in the “matcher” definition (without canceling the `subscribe/2` function, in case someone wants to subscribe to a channel later).\n\nI also noticed that there can be many handlers, why not.\n\nThus, the unique identifier for a matcher remains the match itself. It can be quite tricky, so I chose its textual representation as the id (the string `\"{:tag_answer, _}\"` for the example above). I don’t really like this solution, but I don’t have a better one (yet). At least, inspecting matchers is much easier this way than, for example, using something hash-like.\n\nSo, we get a many-to-many relationship: many handlers for one match, which can receive messages from many channels. For example, we can subscribe to all messages of the form `{:error, _}` from all channels and assign two handlers: a logger will print to the console, telemetry will send something to Grafana, or wherever it’s customary to send everything.","ref":"diy-message-broker.html#pseudocode"},{"type":"extras","title":"Main Architecture - DIY Message Broker","doc":"Back pressure in Elixir implies using the [`GenStage`](https://hexdocs.pm/gen_stage) library. I already implemented my [`Throttler`](https://hexdocs.pm/finitomata/Finitomata.Throttler.html) with it, now it’s the broker’s turn. We remember about horizontal scaling—which means several consumers, at least one per node. Each consumer will send messages to specified channels, matchers will check if they need to bother (whether the message matches), and if so, call the handlers. Sounds reasonable.\n\nI spread consumers across nodes using unnamed processes managed by [`DistributedSupervisor`](https://hexdocs.pm/distributed_supervisor/DistributedSupervisor.html), the broadcaster too (except this process is named, so there’s only one per cluster, and it will be restarted on another node if the current one disappears).\n\nEach matcher is also a process, which stores a list of handlers and the match itself in its state. Here I faced my first architectural dilemma: how to store the match? `{:foo, _}` can’t just be stored as is; such code is only allowed as LHO in direct pattern matching calls, and storing its AST isn’t an option, because then you can’t insert it into a match. So, I decided to generate a matcher function (anyway, implementing `match/4` is only possible as a macro):\n\n```elixir\nquote generated: true, location: :keep do\n  matcher = fn\n    unquote(match) -> true\n    _ -> false\n  end\n  …\nend\n```\n\nGreat. Now, in the matcher process itself, there’s a callback like this:\n\n```elixir\n  @impl GenServer\n  @doc false\n  def handle_cast({:handle_event, channel, event}, state) do\n    if state.matcher.(event) do\n      Enum.each(state.handlers, fn\n        handler when is_function(handler, 1) -> handler.(event)\n        handler when is_function(handler, 2) -> handler.(channel, event)\n        process -> send(process, {:antenna_event, channel, event})\n      end)\n    end\n\n    {:noreply, state}\n  end\n```\n\nAnd we’re ready to send it messages from consumers (matchers are also managed by DistributedSupervisor, i.e., evenly spread across the cluster).\n\nAt this point, I said `git init` because the MVP was shaping up.","ref":"diy-message-broker.html#main-architecture"},{"type":"extras","title":"Synchronous Call - DIY Message Broker","doc":"Up to this point, everything was pretty trivial. But how do you organize a synchronous call when all messages go through seven circles of hell (broadcast, consumers, matchers-all on who-knows-which nodes in the cluster)?\n\nHave you ever wondered why the signature of `GenServer`’s asynchronous callback is of arity `2`, while the synchronous one is of arity `3`? By the way, this is one of my favorite interview questions: it immediately becomes clear whether you’re dealing with a dreamy reactive hipster or a nerdy freaking sociopath.\n\nThe second argument in the [`handle_call/3`](https://hexdocs.pm/elixir/GenServer.html#c:handle_call/3) callback is the process identifier (without loss of generality) waiting for a synchronous response. And instead of the tuple `{:reply, result, state}` from this callback, you can return `{:noreply, new_state}`, and later send a synchronous reply directly to the calling process using [`GenServer.reply/2`](https://hexdocs.pm/elixir/GenServer.html#reply/2). If you didn’t know this, have a sip of whiskey for me: this is the hammer that makes everything around you nails.\n\n`GenStage`, in turn, also exports `reply/2`. So now I just need to pass `from` through all consumers and matchers, and then deep inside write something like:\n\n```elixir\nEnum.each(results, fn\n  {nil, _} -> :ok\n  {from, results} -> GenStage.reply(from, results)\nend)\n```\n\nIf it was an asynchronous call-there’s no `from`, and we do nothing. But if it exists-we send back the accumulated results of all handler calls (and let it choke on them).\n\nThat seems to be all I wanted to share today. The link to the library, its source code, and tests are above.\n\nHappy brokering!","ref":"diy-message-broker.html#synchronous-call"},{"type":"extras","title":"Antenna Cheatsheet","doc":"# Antenna Cheatsheet","ref":"antenna.html"},{"type":"extras","title":"Setting up matchers - Antenna Cheatsheet","doc":"### `match/4`\n\n#### Matches events against a pattern and executes handler when matched\n\n```elixir\nAntenna.match(MyApp.Antenna, pattern, handler, opts \\\\ [])\n```\n\n#### Pattern can be any Elixir pattern with guards\n\n```elixir\n{:user_event, user_id, action}\n%{type: :order, amount: amount} when amount > 1000\n{:error, _reason} = error\n```\n\n#### Handler can be a pid, function, or list of handlers\n\n```elixir\nself()\n&Logger.info/1\nfn channel, event -> handle_event(event) end\n[&Logger.error/1, &Notifier.alert/1]\n```","ref":"antenna.html#setting-up-matchers"},{"type":"extras","title":"Common Patterns - Antenna Cheatsheet","doc":"#### High-value orders\n\n```elixir\nAntenna.match(MyApp.Antenna,\n  %{event: :order_created, total: total} when total > 1000,\n  &HighValueOrderProcessor.handle/1,\n  channels: [:orders])\n```\n\n#### Error monitoring\n\n```elixir\nAntenna.match(MyApp.Antenna, \n  {:error, _},\n  [&Logger.error/1, &Notifier.alert/1],\n  channels: [:system])\n```\n\n#### Temperature alerts\n\n```elixir\nAntenna.match(MyApp.Antenna,\n  {:temperature, celsius, location} when celsius > 30,\n  fn _, event -> handle_high_temp(event) end,\n  channels: [:sensors])\n```","ref":"antenna.html#common-patterns"},{"type":"extras","title":"Handling events - Antenna Cheatsheet","doc":"### `event/3`\n\n#### Sends an event to all the channels specified and responds immediately\n\n```elixir\nAntenna.event(id, channels, event)\n```\n\n#### Broadcasts an event asynchronously to all matching handlers\n\n```elixir\nAntenna.event(MyApp.Antenna, [:users], {:user_login, \"123\"})\nAntenna.event(MyApp.Antenna, [:metrics], %{cpu: 80})\n```\n\n### `sync_event/3,4`\n\n#### Sends an event to all the channels specified and waits for responds\n\n```elixir\nAntenna.sync_event(id, channels, event, timeout \\\\ 5_000)\n```\n\n#### Broadcasts event synchronously and collects responses\n\n```elixir\nresponses = Antenna.sync_event(MyApp.Antenna, [:auth], {:verify, token})\n```","ref":"antenna.html#handling-events"},{"type":"extras","title":"Adding handlers/subscribers to events - Antenna Cheatsheet","doc":"### `subscribe/3`\n\n#### Subscribes the handlers(s) to channel(s)\n\n```elixir\nAntenna.subscribe(id, channels, handler)\n```\n\n#### Subscribe handler to channels manually\n\n```elixir\nAntenna.subscribe(MyApp.Antenna, [:logs], &Logger.info/1)\nAntenna.subscribe(MyApp.Antenna, [:errors], error_handler_pid)\n```","ref":"antenna.html#adding-handlers-subscribers-to-events"},{"type":"extras","title":"Testing - Antenna Cheatsheet","doc":"","ref":"antenna.html#testing"},{"type":"extras","title":"Match and receive events - Antenna Cheatsheet","doc":"#### Declaration\n\n```elixir\ntest \"matches events\" do\n  Antenna.match(TestAntenna,\n    {:test, value} when value > 0,\n    self(),\n    channels: [:test])\n  # ⇓ see below ⇓ \n```\n\n#### Validation\n\n```elixir\n  # ⇑ see above ⇑ \n  Antenna.event(TestAntenna, [:test], {:test, 42})\n  assert_receive {:antenna_event, :test, {:test, 42}}\nend\n```","ref":"antenna.html#match-and-receive-events"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.2"}}